/*******************************************************************************
*    DynDNSPalm 1.0 for PalmOS                                                 *
*    Copyright (C) 2006-2007 Luca Bertoncello                                  *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <PalmOS.h>
#include <PalmTypes.h>
#include <PalmCompatibility.h>
#include <ConnectionMgr.h>
#include <SystemMgr.h>
#include <NetMgr.h>
#include <Unix/sys_socket.h>
#include <netinet_in.h>
#include "dyndnsPalm.h"
#include "prefs.h"

#define appID                   'DDP'
#define versionID               1

static FormPtr gpForm;
UInt16 AppNetRefnum, Instance;
Preferences gAppPrefs;

static Err StartApplication() 
{
  FrmGotoForm(MainForm);
  return 0;
}

static void StopApplication() 
{
  FrmCloseAllForms();
}

Char* GetField(UInt16 formID, UInt16 fieldID)
{
  FormPtr frm;
  FieldPtr fld;
  Char *p;

  frm = FrmGetFormPtr(formID);
  fld = (FieldPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  p = FldGetTextPtr(fld);

  return (p ? p : "");
}

void SetField(UInt16 formID, UInt16 fieldID, MemPtr str)
{
  FormPtr frm;
  FieldPtr fld;
  CharPtr p;
  VoidHand h;

  frm = FrmGetFormPtr(formID);
  fld = (FieldPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  h = (VoidHand) FldGetTextHandle(fld);
  if(h == NULL)
  {
    h = MemHandleNew(FldGetMaxChars(fld)+1);
    ErrFatalDisplayIf(!h, "No Memory");
  }

  p = (CharPtr) MemHandleLock(h);
  StrCopy(p, str);
  MemHandleUnlock(h);
  FldSetTextHandle(fld, (Handle) h);
}

Char encode(Char u)
{
  if(u < 26)  return 'A' + u;
  if(u < 52)  return 'a' + (u - 26);
  if(u < 62)  return '0' + (u - 52);
  if(u == 62) return '+';
  
  return '/';
}

Char *encode_base64(Char *src)
{
  int i, p, size;
  Char *out;

  if(!src)
    return NULL;

  size = StrLen(src);

  out = MemPtrNew((size * 4 / 3 +4) * sizeof(Char));
  MemSet(out, (size * 4 / 3 + 4), 0);
  p = 0;

  for(i = 0; i < size; i += 3)
  {
    Char b1=0, b2=0, b3=0, b4=0, b5=0, b6=0, b7=0;

    b1 = src[i];

    if(i + 1 < size)
      b2 = src[i + 1];

    if(i + 2 < size)
      b3 = src[i + 2];

    b4 = b1>>2;
    b5 = ((b1 & 0x3) << 4) | (b2 >> 4);
    b6 = ((b2 & 0xf) << 2) | (b3 >> 6);
    b7 = b3 & 0x3f;

    out[p++] = encode(b4);
    out[p++] = encode(b5);

    if(i + 1 < size)
      out[p++] = encode(b6);
    else
      out[p++] = '=';

    if(i + 2 < size)
      out[p++] = encode(b7);
    else
      out[p++] = '=';
  }

  return out;
}

void doDynDNS(int type)
{
  Char *answer, str[255], dynDnsStatus[255], *userPass;
  Char username[255], password[255], hostname[255];
  FormPtr frm = FrmGetFormPtr(MainForm);
  NetSocketRef socket;
  Err error = errNone;
  NetHostInfoPtr host;
  NetHostInfoBufPtr buf;
  NetSocketAddrINType sockaddr;
  ULong **addrp;
  UInt16 nBytes, l, i;
  Boolean hasToGet, commProb;
  Int32 AppNetTimeout;

  if(SysLibFind("Net.lib", &AppNetRefnum) == 0)
  {
    NetLibOpen(AppNetRefnum, &Instance);

    commProb = false;
    AppNetTimeout = SysTicksPerSecond() * 15; // TimeOut: 15 seconds
    StrCopy(username, GetField(MainForm, idUsername));
    StrCopy(password, GetField(MainForm, idPassword));
    StrCopy(hostname, GetField(MainForm, idHostname));
    StrPrintF(str, "%s:%s", username, password);
    userPass = encode_base64(str);
    answer = MemPtrNew(1024 * sizeof(Char));
    MemSet(answer, 1024, 0);
    StrCopy(str, "Connecting to dyndns...");
    SetField(MainForm, idStatus, str);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));

    socket = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET, netSocketTypeStream, netSocketProtoIPTCP, AppNetTimeout, &error);
    if(!error)
    {
      buf = (NetHostInfoBufPtr)MemPtrNew(sizeof(NetHostInfoBufType));
      host = NetLibGetHostByName(AppNetRefnum, "members.dyndns.org", buf, AppNetTimeout, &error);
      if(host != 0)
      {
        sockaddr.family = AF_INET;
        sockaddr.port = htons(80);
        sockaddr.addr = INADDR_ANY;
        for(addrp = (ULong **)buf->hostInfo.addrListP; *addrp; ++addrp)
        {
          ULong a = htonl(**addrp);
          if(a != INADDR_ANY)
          {
            sockaddr.addr = a;
            break;
          }
        }
        MemPtrFree(buf);
        if(NetLibSocketConnect(AppNetRefnum, socket, (NetSocketAddrType*)&sockaddr, sizeof(NetSocketAddrType), AppNetTimeout, &error) != -1)
        {
          StrCopy(str, "Sending host information...");
          SetField(MainForm, idStatus, str);
          FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));

          if(type == 0)
            StrPrintF(str, "GET /nic/update?system=dyndns&hostname=%s&myip=127.0.0.1 HTTP/1.0", hostname);
          else
            StrPrintF(str, "GET /nic/update?system=dyndns&hostname=%s HTTP/1.0", hostname);
          NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
          NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
          StrCopy(str, "User-Agent: DynDNSPalm 1.0");
          NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
          NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
          StrCopy(str, "Host: members.dyndns.org");
          NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
          NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
          StrCopy(str, "Accept: */*");
          NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
          NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
          StrCopy(str, "Connection: Keep-Alive");
          NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
          NetLibSend(AppNetRefnum, socket, "\r\n", 2, 0x0, NULL, 0, AppNetTimeout, &error);
          StrPrintF(str, "Authorization: Basic %s", userPass);
          NetLibSend(AppNetRefnum, socket, str, StrLen(str), 0x0, NULL, 0, AppNetTimeout, &error);
          NetLibSend(AppNetRefnum, socket, "\r\n\r\n", 4, 0x0, NULL, 0, AppNetTimeout, &error);
          nBytes = 1;
          while(nBytes != 0)
          {
            MemSet(str, 255, 0);
            nBytes = NetLibReceive(AppNetRefnum, socket, str, 250, 0x0, NULL, NULL, AppNetTimeout, &error);
            if(nBytes != 0)
              StrCat(answer, str);
          }
          hasToGet = false;
          MemSet(str, 255, 0);
          MemSet(dynDnsStatus, 255, 0);
          for(i = l = 0; l < StrLen(answer); l++)
          {
            if(answer[l] == '\n')
            {
              l++;
              i = 0;
              MemSet(str, 255, 0);
            }
            if(answer[l] != '\r' && answer[l] != '\n')
              str[i++] = answer[l];
            else
            {
              if(hasToGet == true && StrCompare(str, "") != 0)
                StrPrintF(dynDnsStatus, "DynDNS says: [%s]", str);
              if(StrCompare(str, "Connection: close") == 0)
                hasToGet = true;
            }
          }
          if(hasToGet == true && StrCompare(str, "") != 0)
            StrPrintF(dynDnsStatus, "DynDNS: [%s]", str);
          SetField(MainForm, idStatus, dynDnsStatus);
          FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));

          NetLibSocketClose(AppNetRefnum, socket, AppNetTimeout, &error);
        }
        else
          commProb = true;
      }
      else
        commProb = true;
    }
    else
      commProb = true;

    if(commProb == true)
    {
      StrCopy(str, "Unable to communicate with dyndns!");
      SetField(MainForm, idStatus, str);
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idStatus)));
    }

    MemPtrFree(answer);
    MemPtrFree(userPass);
    NetLibClose(AppNetRefnum, false);
  }
}

Boolean MainFormEventHandler(EventPtr event)
{
  Boolean handled = false;
  FormPtr frm = FrmGetFormPtr(MainForm);
  FieldPtr fld;
  UInt16 nAppPrefsSize;

  switch(event->eType)
  {
    case frmOpenEvent:
      FrmDrawForm(FrmGetFormPtr(MainForm));
      nAppPrefsSize = sizeof(Preferences);
      if(PrefGetAppPreferences(appID, 1, &gAppPrefs, &nAppPrefsSize, true) == noPreferenceFound)
        MemSet(&gAppPrefs, sizeof(Preferences), 0);
      CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword)), gAppPrefs.hasToDisplayPassword);
      if(CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword))))
        FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), stdFont);
      else
        FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), symbol11Font);
      SetField(MainForm, idUsername, gAppPrefs.username);
      SetField(MainForm, idPassword, gAppPrefs.password);
      SetField(MainForm, idHostname, gAppPrefs.hostname);
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idUsername)));
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)));
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idHostname)));
      handled = true;
    break;

    case frmCloseEvent:
      StrCopy(gAppPrefs.username, GetField(MainForm, idUsername));
      StrCopy(gAppPrefs.password, GetField(MainForm, idPassword));
      StrCopy(gAppPrefs.hostname, GetField(MainForm, idHostname));
      gAppPrefs.hasToDisplayPassword = CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword)));
      PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);
      FrmEraseForm(FrmGetFormPtr(MainForm));
      FrmDeleteForm(FrmGetFormPtr(MainForm));
      handled = true;
    break;

    case ctlSelectEvent:
      switch (event->data.ctlSelect.controlID)
      {
        case displayPassword:
          if(CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword))))
            FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), stdFont);
          else
            FldSetFont(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)), symbol11Font);
          FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idPassword)));
          handled = true;
        break;
        case idButtonSet:
          StrCopy(gAppPrefs.username, GetField(MainForm, idUsername));
          StrCopy(gAppPrefs.password, GetField(MainForm, idPassword));
          StrCopy(gAppPrefs.hostname, GetField(MainForm, idHostname));
          gAppPrefs.hasToDisplayPassword = CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword)));
          PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);
          doDynDNS(1);
          handled = true;
        break;
        case idButtonReset:
          StrCopy(gAppPrefs.username, GetField(MainForm, idUsername));
          StrCopy(gAppPrefs.password, GetField(MainForm, idPassword));
          StrCopy(gAppPrefs.hostname, GetField(MainForm, idHostname));
          gAppPrefs.hasToDisplayPassword = CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, displayPassword)));
          PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);
          doDynDNS(0);
          handled = true;
        break;
      }
    break;

    default:
    break;
  }

  return handled;
}

static Boolean ApplicationEventHandler(EventPtr event) 
{
  Boolean handled = false;
  FormType* frmP;
  ListType *lstP;

  switch(event->eType)
  {
    case frmLoadEvent:
    {
      FormPtr pForm = FrmInitForm(event->data.frmLoad.formID);
      FrmSetActiveForm(pForm);
      switch(event->data.frmLoad.formID)
      {
        case MainForm:
          FrmSetEventHandler(pForm, MainFormEventHandler);
        break;

        default:
        break;
      }
      handled = true;
    }
    break;

    case menuEvent:
      switch(event->data.menu.itemID)
      {
      	case HelpMenuAbout:
          FrmAlert(AboutAlert);
        break;
      }
      handled = true;
    break;

    default:
    break;
  }

  return handled;
}

static void EventLoop() 
{
  EventType event;
  UInt16 error;

  do
  {
    EvtGetEvent(&event, evtWaitForever);

    if (!SysHandleEvent(&event))
    {
      if (!MenuHandleEvent(0, &event, &error))
      {
        if (!ApplicationEventHandler(&event))
          FrmDispatchEvent(&event);
      }
    }
  } while(event.eType != appStopEvent);
}

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) 
{
  Err err = 0;
  UInt32 romVersion;

  switch(launchCode)
  {
    case sysAppLaunchCmdNormalLaunch:
    {
      FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
      if(romVersion < sysMakeROMVersion(3, 5, 0, sysROMStageRelease, 0))
      {
        FrmAlert(RomIncompatibleAlert);
        return sysErrRomIncompatible;
      }
      if((err = StartApplication()) == 0)
      {
        EventLoop();
        StopApplication();
      }
      break;
    }
  }

  return err;
}
